#include <Wire.h>
#include <math.h>
#include <BMI160Gen.h>
#include <LowPower.h>
#include <RTC.h>
#include <Arduino.h>
#include <File.h>
#include <Flash.h>
#include "bp35c0-j11.h"

#define BAUDRATE  115200  //115200
#define SENSE_RATE   200
#define GYRO_RANGE   250
#define ACCL_RANGE     2
#define SLEEPTIME      3

#define COUNT_FOR_CALIB 1024

#define deg_to_rad(a) (a/180*M_PI)
#define rad_to_deg(a) (a/M_PI*180)

#define PRINT_ACCL

uint16_t adjust_usec;
static float calib_accel_x = 0, calib_accel_y = 0, calib_accel_z = 0;
volatile int alarmInt      = 0;
const int LEDPIN           = 23;                //LED pin number for alert
const double threthold     = 5.0;        //***Threshold for alerting (unit:degree)*******
unsigned char state = 0;
char machineID = "ID:LSA0001";

BP35C0J11 bp35c0j11;

char * dtostrf(double value, unsigned int width, unsigned int decimalPlaces, char* buf)
{
  char fmt[20];

  snprintf(fmt, 20, "%%%d.%df", width, decimalPlaces);
  sprintf(buf, fmt, value);
  return buf;
}

void printClockMode()
{
  clockmode_e mode = LowPower.getClockMode();

  Serial.println("--------------------------------------------------");
  Serial.print("clock mode: ");
  switch (mode) {
  case CLOCK_MODE_156MHz: Serial.println("156MHz"); break;
  case CLOCK_MODE_32MHz:  Serial.println("32MHz"); break;
  case CLOCK_MODE_8MHz:   Serial.println("8MHz"); break;
  }
}

float convertRawAccel(int aRaw) 
{
  // ex) if the range is +/-2g ; +/-32768/2 = +/-16384 LSB/g
  float lsb_g = float(0x7FFF) / ACCL_RANGE;
  return aRaw / lsb_g;
}


void setup() {
  Serial.begin(BAUDRATE);
  RTC.begin();
  LowPower.begin();
  LowPower.clockMode(CLOCK_MODE_32MHz); //156,32,8
//  printClockMode();
  bootcause_e bc = LowPower.bootCause();
  Serial.print("bc = ");
  Serial.println(bc);
  pinMode(LEDPIN, OUTPUT);
  digitalWrite(LEDPIN, LOW);

  digitalWrite(LED0,LOW);
  digitalWrite(LED1,LOW);
  digitalWrite(LED2,LOW);
  digitalWrite(LED3,LOW);

  if (bc == 0){
    digitalWrite(LEDPIN,HIGH);
    digitalWrite(LED3,HIGH);
    delay(10);
    digitalWrite(LEDPIN,LOW);
    digitalWrite(LED3,LOW);
  }
  boolean rc = FALSE ;
  bp35c0j11.j11_init();
  rc = bp35c0j11.wait_msg();
  if(rc == TRUE){
    state = 1 ;           //  hardware reset end 
  }else{
    state = 0 ;
  }

  while(state < 8){
    delay(500);
    unsigned char msg_length = 0 ;
    boolean rc = 0 ;
    Serial.print("State = ");
    Serial.println(state, DEC);
    delay(500);
    switch (state) {
      case(0):   // need hardware reset 
        rc = bp35c0j11.cmd_send(CMD_RESET);
        rc = bp35c0j11.wait_msg();
        if(rc == TRUE){
          state = 1 ;
        }
        break;
      case(1):  //  init state 
        rc = bp35c0j11.cmd_send(CMD_INI);
        rc = bp35c0j11.wait_msg();
        if(rc == TRUE){
          state = 2;
        }
        break;
      case(2): // HAN PANA setting
        rc = bp35c0j11.cmd_send(CMD_PANA_SET);
        rc = bp35c0j11.wait_msg();
        if(rc == TRUE){
          state = 4;
        }
        break;
      case(3): // active scan
        rc = bp35c0j11.cmd_send(CMD_SCAN);
        rc = bp35c0j11.wait_msg();
        if(rc == TRUE){
          rc = bp35c0j11.wait_msg();
          if(rc == TRUE){
            state = 4;
          }
        }
        break;
      case(4): //  HAN act
        rc = bp35c0j11.cmd_send(CMD_HAN);
        rc = bp35c0j11.wait_msg();
        if(rc == TRUE){
          state = 5;
        }
        break;
      case(5): // HAN PANA act
        rc = bp35c0j11.cmd_send(CMD_PANA);
        rc = bp35c0j11.wait_msg();
        if(rc == TRUE){
          rc = bp35c0j11.wait_msg();
          if(rc == TRUE){
            state = 6;
          }
        }
        break;
      case(6): // rcv mode change
        rc = bp35c0j11.cmd_send(CMD_CON_SET);
        rc = bp35c0j11.wait_msg();
        if(rc == TRUE){
          state = 7;
        }
        break;
      case(7): // my_port open
        rc = bp35c0j11.cmd_send(CMD_PORTOPEN);
        rc = bp35c0j11.wait_msg();
        if(rc == TRUE){
          state = 8;
        }
        break;
    }
  }

  if (bc == 0){
    digitalWrite(LEDPIN,HIGH);
    digitalWrite(LED3,HIGH);
    delay(10);
    digitalWrite(LEDPIN,LOW);
    digitalWrite(LED3,LOW);
    delay(300);
    digitalWrite(LEDPIN,HIGH);
    digitalWrite(LED3,HIGH);
    delay(10);
    digitalWrite(LEDPIN,LOW);
    digitalWrite(LED3,LOW);
  }
  
  BMI160.begin(BMI160GenClass::I2C_MODE);
  BMI160.setAccelerometerRate(SENSE_RATE);
  BMI160.setAccelerometerRange(ACCL_RANGE);

  // Serial.println("Calibrating...");
  unsigned long sampling_rate = 1000 / SENSE_RATE;

  // Calcurate offset
  for (int i = 0; i < COUNT_FOR_CALIB; ++i) {
    int xAcc, yAcc, zAcc;
    BMI160.readAccelerometer(xAcc, yAcc, zAcc);
    calib_accel_x += convertRawAccel(xAcc);
    calib_accel_y += convertRawAccel(yAcc);
    calib_accel_z += convertRawAccel(zAcc);
  }
  
  // mean value
  calib_accel_x /= COUNT_FOR_CALIB;
  calib_accel_y /= COUNT_FOR_CALIB;
  calib_accel_z /= COUNT_FOR_CALIB;
  double xInit = calib_accel_x, yInit = calib_accel_y, zInit = calib_accel_z;
  
  // Processing at initial startup
  if (bc == 0){
    if (Flash.exists("initData.txt")) Flash.remove("initData.txt");
    File initData = Flash.open("initData.txt",FILE_WRITE);
    initData.print(String(calib_accel_x, 3));
    initData.print(",");
    initData.print(String(calib_accel_y, 3));
    initData.print(",");
    initData.print(String(calib_accel_z, 3));
    initData.print(",");
    initData.close();
    bp35c0j11.radiodata = "ID:LSA0001";
    rc = bp35c0j11.cmd_send(CMD_UDPSEND);
    rc = bp35c0j11.wait_msg();
    bp35c0j11.radiodata = "STATUS:launch";
    rc = bp35c0j11.cmd_send(CMD_UDPSEND);
    rc = bp35c0j11.wait_msg();  
    bp35c0j11.radiodata = "GNSS:1111/2222";
    rc = bp35c0j11.cmd_send(CMD_UDPSEND);
    rc = bp35c0j11.wait_msg();
 
  }

  //Processing after deep sleep
  int counter = 0;
  if (bc == 5){
    String str;
    File initData = Flash.open("initData.txt");  
    while(initData.available()){
      if (counter == 0) {
        xInit = initData.readStringUntil(',').toDouble();
      }else if (counter == 1){
        yInit = initData.readStringUntil(',').toDouble();
      }else if (counter == 2){
        zInit = initData.readStringUntil(',').toDouble();
      }else{
        str += char(initData.read());
      }
      counter++;
    }
    initData.close();
  }

  // SENSE_RATE is Hz and 2 millsec is the process time for this sketch
  adjust_usec = (1000/SENSE_RATE - 2) * 1000; 

  // Calculate the angle between the initial angle and the current angle
  double numerator = (xInit * calib_accel_x) + (yInit * calib_accel_y) + (zInit * calib_accel_z);
  double denominator1 = sqrt(pow(abs(xInit),2) + pow(abs(yInit),2) + pow(abs(zInit),2));
  double denominator2 = sqrt(pow(abs(calib_accel_x),2) + pow(abs(calib_accel_y),2) + pow(abs(calib_accel_z),2));
  double degree = acos(numerator/(denominator1*denominator2))*(180/3.141592654);
//  Serial.println(degree);

  if (bc == 5){
    bp35c0j11.radiodata = "ID:LSA0001";
    rc = bp35c0j11.cmd_send(CMD_UDPSEND);
    rc = bp35c0j11.wait_msg();

  //alart output setting
    if (degree >= threthold){
  //  if (degree <0){
      bp35c0j11.radiodata = "STATUS:alert";
      rc = bp35c0j11.cmd_send(CMD_UDPSEND);
      rc = bp35c0j11.wait_msg();
      bp35c0j11.radiodata = "DEGREE:";
      char chardegree[7];
      dtostrf(degree,3,3,chardegree);
      strcat(bp35c0j11.radiodata, chardegree);
      rc = bp35c0j11.cmd_send(CMD_UDPSEND);
      rc = bp35c0j11.wait_msg();
    
      while(true){
        digitalWrite(LEDPIN,HIGH);
        digitalWrite(LED3,HIGH);
        delay(10);
        digitalWrite(LEDPIN,LOW);
        digitalWrite(LED3,LOW);
        delay(1000);
      }
    }
    else {
      bp35c0j11.radiodata = "STATUS:in range";
      rc = bp35c0j11.cmd_send(CMD_UDPSEND);
      rc = bp35c0j11.wait_msg();
      bp35c0j11.radiodata = "DEGREE:";
      char chardegree[7];
      dtostrf(degree,3,3,chardegree);
      strcat(bp35c0j11.radiodata, chardegree);
      rc = bp35c0j11.cmd_send(CMD_UDPSEND);
      rc = bp35c0j11.wait_msg();  
    }
  }
  LowPower.deepSleep(SLEEPTIME);
}

void loop() {

}
